/**
 * Copyright (c) Members of the EGEE Collaboration. 2007-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *
 */

#define _XOPEN_SOURCE	500

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <sys/types.h>

#include <pthread.h>

#include "xacml_attr_obl_profile.h"
#include "pdp_xacml_authz_process.h"


#include "xacml_server.h"
#include "xacml_io_ssl.h"


#include "pdp_xacml_lcas_lcmaps.h"

#include "scas_log.h"

#include <time.h>




/* impl */



/************************************************
Function:    extractRequestSubject
Parameters:  const xacml_request_t           request
             xacml_profile_request_t * xacml_profile_request
Description:
             This function extracts the Subject information details
             from the request and puts that into the xacml_profile_request_t 
             object.
Return:
          0 : good
         !0 : bad 
************************************************/
int extractRequestSubject (const xacml_request_t           request,
                                 xacml_profile_request_t * xacml_profile_request)
{
    char * logstr = "extractRequestSubject()";
    size_t count = 0;
    size_t i     = 0;

    xacml_request_get_subject_attribute_count(request, &count);

    for (i = 0; i < count; i++)
    {
        const char *category;
        const char *attribute_id;
        const char *data_type;
        const char *issuer;
        const char *value;

        xacml_request_get_subject_attribute(
                request, i, &category, &attribute_id, &data_type, &issuer, &value);

        /* Should only be send once */
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_SUBJECT_X509_ID) == 0)                    
        {
            if (!(xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_id.subject_x509_id                             = strdup (value)))
            { 
                scas_log(LOG_ERR, "%s: Failure in allocating memory @  1\n", logstr);
                return -1;
            }
            continue;
        }
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_SUBJECT_X509_ISSUER) == 0)                
        {
            if (!(xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_issuer.subject_x509_issuer                     = strdup (value)))
            { 
                scas_log(LOG_ERR, "%s: Failure in allocating memory @  2\n", logstr);
                return -1;
            }
            continue;
        }
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_SUBJECT_CERTIFICATE_SERIAL_NUMBER) == 0)  
        {
            xacml_profile_request -> xacml_profile_subject.xacml_subject_certificate_serial_number.certificate_serial_number = atoi (value);
            /* Should handle atoi() failures */
            continue;
        }
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_CERT_CHAIN) == 0)
        {
            if (!(xacml_profile_request -> xacml_profile_subject.xacml_cert_chain.cert_chain                                       = strdup (value)))
            { 
                scas_log(LOG_ERR, "%s: Failure in allocating memory @  3\n", logstr);
                return -1;
            }
            continue;
        }
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_SUBJECT_VOMS_PRIMARY_FQAN) == 0)
        {
            if (!(xacml_profile_request -> xacml_profile_subject.xacml_subject_primary_voms_fqan.subject_primary_voms_fqan         = strdup (value)))
            { 
                scas_log(LOG_ERR, "%s: Failure in allocating memory @  4\n", logstr);
                return -1;
            }
            continue;
        }


        /* Multiple elements */
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_SUBJECT_CA_SERIAL_NUMBER) == 0)
        {
            xacml_profile_request -> xacml_profile_subject.xacml_ca_serial_number_cnt++;
            if (!(xacml_profile_request -> xacml_profile_subject.xacml_ca_serial_number = realloc (xacml_profile_request -> xacml_profile_subject.xacml_ca_serial_number, 
                                                                                                   sizeof (xacml_ca_serial_number_t) * (xacml_profile_request -> xacml_profile_subject.xacml_ca_serial_number_cnt))))
            {
                /* Restore of the counter, needed for the clean up function */
                xacml_profile_request -> xacml_profile_subject.xacml_ca_serial_number_cnt--;
                scas_log(LOG_ERR, "%s: Failure in allocating memory @  5\n", logstr);
                return -1;
            }

            xacml_profile_request -> xacml_profile_subject.xacml_ca_serial_number[xacml_profile_request -> xacml_profile_subject.xacml_ca_serial_number_cnt - 1].ca_serial_number = atoi(value);
            continue;
        }
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_CA_POLICY_OID) == 0)
        {
            xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid_cnt++;
            if (!(xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid = realloc (xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid, 
                                                                                                sizeof (xacml_ca_policy_oid_t) * xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid_cnt)))
            {
                /* Restore of the counter, needed for the clean up function */
                xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid_cnt--;   
                scas_log(LOG_ERR, "%s: Failure in allocating memory @  6\n", logstr);
                return -1;
            }

            if (!(xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid[xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid_cnt - 1].ca_policy_oid          = strdup (value)))
            {
                scas_log(LOG_ERR, "%s: Failure in allocating memory @  7\n", logstr);
                return -1;
            }
            continue;
        }
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_VO) == 0)
        {
            xacml_profile_request -> xacml_profile_subject.xacml_subject_vo_cnt++;
            if (!(xacml_profile_request -> xacml_profile_subject.xacml_subject_vo = realloc (xacml_profile_request -> xacml_profile_subject.xacml_subject_vo, 
                                                                                             sizeof (xacml_subject_vo_t) * xacml_profile_request -> xacml_profile_subject.xacml_subject_vo_cnt)))
            {
                scas_log(LOG_ERR, "%s: Failure in allocating memory @  8\n", logstr);
                /* Restore of the counter, needed for the clean up function */
                xacml_profile_request -> xacml_profile_subject.xacml_subject_vo_cnt--;
                return -1;
            }

            if (!(xacml_profile_request -> xacml_profile_subject.xacml_subject_vo[xacml_profile_request -> xacml_profile_subject.xacml_subject_vo_cnt - 1].subject_vo = strdup (value)))
            {
                scas_log(LOG_ERR, "%s: Failure in allocating memory @  9\n", logstr);
                return -1;
            }
            continue;
        }
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_VOMS_SIGNING_SUBJECT) == 0)
        {
            xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject_cnt++;
            if (!(xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject = realloc (xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject, 
                                                                                                       sizeof (xacml_voms_signing_subject_t) * xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject_cnt)))
            {
                /* Restore of the counter, needed for the clean up function */
                scas_log(LOG_ERR, "%s: Failure in allocating memory @ 10\n", logstr);
                xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject_cnt--;
                return -1;
            }

            if (!(xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject[xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject_cnt - 1].voms_signing_subject = strdup (value)))
            {
                scas_log(LOG_ERR, "%s: Failure in allocating memory @ 11\n", logstr);
                return -1;
            }
            continue;
        }
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_VOMS_SIGNING_ISSUER) == 0)
        {
            xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer_cnt++;
            if (!(xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer = realloc (xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer, 
                                                                                                      sizeof (xacml_voms_signing_issuer_t) * xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer_cnt)))
            {
                /* Restore of the counter, needed for the clean up function */
                xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer_cnt--;
                scas_log(LOG_ERR, "%s: Failure in allocating memory @ 12\n", logstr);
                return -1;
            }

            if (!(xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer[xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer_cnt - 1].voms_signing_issuer = strdup (value)))
            {
                scas_log(LOG_ERR, "%s: Failure in allocating memory @ 13\n", logstr);
                return -1;
            }
            continue;
        }
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_VOMS_DNS_PORT) == 0)
        {
            xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port_cnt++;
            if (!(xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port = realloc (xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port, 
                                                                                                sizeof (xacml_voms_dns_port_t) * xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port_cnt)))
            {
                /* Restore of the counter, needed for the clean up function */
                xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port_cnt--;
                scas_log(LOG_ERR, "%s: Failure in allocating memory @ 14\n", logstr);
                return -1;
            }

            if (!(xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port[xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port_cnt - 1].voms_dns_port = strdup (value)))
            {
                scas_log(LOG_ERR, "%s: Failure in allocating memory @ 15\n", logstr);
                return -1;
            }  
            continue;
        }
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_SUBJECT_VOMS_FQAN) == 0)
        {
            xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan_cnt++;
            if (!(xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan = realloc (xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan, 
                                                                                                    sizeof (xacml_subject_voms_fqan_t) * xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan_cnt)))
            
            {
                /* Restore of the counter, needed for the clean up function */
                xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan_cnt--;
                scas_log(LOG_ERR, "%s: Failure in allocating memory @ 16\n", logstr);
                return -1;
            }  

            if (!(xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan[xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan_cnt - 1].subject_voms_fqan = strdup (value)))
            {
                scas_log(LOG_ERR, "%s: Failure in allocating memory @ 17\n", logstr);
                return -1;
            }  
            continue;
        }



        if (strcmp(category, XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT))
        {
            continue;
        }
        if (strcmp(attribute_id, XACML_SUBJECT_ATTRIBUTE_SUBJECT_ID))
        {
            continue;
        }
    }

    return 0;
}



/************************************************
Function:    extractRequestAction
Parameters:  const xacml_request_t           request
             xacml_profile_request_t * xacml_profile_request
Description:
             This function extracts the Action information details
             from the request and puts that into the xacml_profile_request_t 
             object.
Return:
          0 : good
         !0 : bad 
************************************************/
int extractRequestAction (const xacml_request_t           request,
                                xacml_profile_request_t * xacml_profile_request)
{
    xacml_result_t                      rc;
    size_t                              i;
    size_t                              count;
    const char *                        attribute_id;
    const char *                        data_type;
    const char *                        issuer;
    const char *                        value;
    const char *                        logstr = "extractRequestAction";

    rc = xacml_request_get_action_attribute_count(request, &count);
    if (rc!=XACML_RESULT_SUCCESS)	{
	scas_log(LOG_ERR, "%s: xacml_request_get_action_attribute_count() failed\n",logstr);
	return -1;
    }


    for (i = 0 ; i < count; i++)
    {
        rc = xacml_request_get_action_attribute(
                request,
                i,
                &attribute_id,
                &data_type,
                &issuer,
                &value);
	if (rc!=XACML_RESULT_SUCCESS)	{
	    scas_log(LOG_ERR, "%s: xacml_request_get_action_attribute() failed\n",logstr);
	    return -1;
	}

        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_ID) == 0)
        {
            if (strcmp (value, XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_ACCESS_FILE) == 0)
            {
                xacml_profile_request -> xacml_profile_action.xacml_action_job_type  = ACTION_ACCESSFILE;

            }
            else if (strcmp (value, XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_QUEUE) == 0) 
            {
                xacml_profile_request -> xacml_profile_action.xacml_action_job_type = ACTION_QUEUE;

            }
            else if (strcmp (value, XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_EXEC_NOW) == 0) 
            {
                xacml_profile_request -> xacml_profile_action.xacml_action_job_type = ACTION_EXECUTE_NOW;

            }
            else
            { 
                scas_log(LOG_NOTICE, "%s: Unsupported Action-id '%s' found: %s\n", logstr, attribute_id, value);
            }
        }
        else if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_RSL_STRING) == 0)
        {
            if (!(xacml_profile_request -> xacml_profile_action.xacml_action_rsl_string.action_rsl_string = strdup(value)))
            {
                scas_log(LOG_ERR, "%s: Failure in allocating memory @ 1\n", logstr);
                return -1;
            } 
        }
        else
        { 
            scas_log_debug (5, "%s: Unknown action attribute-id '%s' found with value %s\n", logstr, attribute_id, value);
        }
    }

    return 0;
}



/************************************************
Function:    extractRequestResource
Parameters:  const xacml_request_t           request
             xacml_profile_request_t * xacml_profile_request
Description:
             This function extracts the Action information details
             from the request and puts that into the xacml_profile_request_t 
             object.
Return:
          0 : good
         !0 : bad 
************************************************/
int extractRequestResource (const xacml_request_t           request,
                                  xacml_profile_request_t * xacml_profile_request)
{
    xacml_result_t                      rc;
    size_t                              i;
    size_t                              j;
    size_t                              count;
    const char *                        attribute_id;
    const char *                        data_type;
    const char *                        issuer;
    const char *                        value;
    xacml_resource_attribute_t          resource_attribute  = NULL;
    size_t                              resource_attr_count = 0;
    const char *                        logstr = "extractRequestResource";


    rc = xacml_request_get_resource_attribute_count(request, &resource_attr_count);
    if (rc!=XACML_RESULT_SUCCESS)	{
	scas_log(LOG_ERR, "%s: xacml_request_get_resource_attribute_count() failed\n",logstr);
	return -1;
    }

    for (i = 0 ; i < resource_attr_count; i++)
    {
        rc = xacml_request_get_resource_attribute(
                request,
                i,
                &resource_attribute);
	if (rc!=XACML_RESULT_SUCCESS)	{
	    scas_log(LOG_ERR, "%s: xacml_request_get_resource_attribute() failed\n",logstr);
	    return -1;
	}


        if (resource_attribute)
        {
            rc = xacml_resource_attribute_get_count(
                       resource_attribute,
                       &count);
	    if (rc!=XACML_RESULT_SUCCESS)	{
		scas_log(LOG_ERR, "%s: xacml_resource_attribute_get_count() failed\n",logstr);
		return -1;
	    }

            for (j = 0; j < count; j++)
            {
                rc = xacml_resource_attribute_get_attribute(
                           resource_attribute,
                           j,
                           &attribute_id,
                           &data_type,
                           &issuer,
                           &value);
		if (rc!=XACML_RESULT_SUCCESS)	{
		    scas_log(LOG_ERR, "%s: xacml_resource_attribute_get_attribute() failed\n",logstr);
		    return -1;
		}

                if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_ID) == 0)
                {
                    if (strcmp (value, XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_CE) == 0)
                    {
                        xacml_profile_request -> xacml_profile_resource.xacml_resource_node_type = RESOURCE_CE;

                    }
                    else if (strcmp (value, XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_SE) == 0)
                    {
                        xacml_profile_request -> xacml_profile_resource.xacml_resource_node_type = RESOURCE_SE;

                    }
                    else if (strcmp (value, XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_WN) == 0)
                    {
                        xacml_profile_request -> xacml_profile_resource.xacml_resource_node_type = RESOURCE_WN;

                    }
                    else if (strcmp (value, XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_RB) == 0)
                    {
                        xacml_profile_request -> xacml_profile_resource.xacml_resource_node_type = RESOURCE_RB;

                    }
                }
                else if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_X509_ID) == 0)
                {
                    if (!(xacml_profile_request -> xacml_profile_resource.xacml_resource_x509_id.resource_x509_id = strdup(value)))
                    {
                        scas_log(LOG_ERR, "%s: Failure in allocating memory @ 1\n", logstr);
                        return -1;
                    } 
                }
                else if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_X509_ISSUER) == 0)
                {
                    if (!(xacml_profile_request -> xacml_profile_resource.xacml_resource_x509_issuer.resource_x509_issuer = strdup(value)))
                    {
                        scas_log(LOG_ERR, "%s: Failure in allocating memory @ 2\n", logstr);
                        return -1;
                    } 
                }
                else if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_RESOURCE_DNS_HOST_NAME) == 0)
                {
                    if (!(xacml_profile_request -> xacml_profile_resource.xacml_resource_dns_name.resource_dns_name = strdup(value)))
                    {
                        scas_log(LOG_ERR, "%s: Failure in allocating memory @ 3\n", logstr);
                        return -1;
                    } 
                }
                else
                { 
                    scas_log_debug (5, "%s: Unknown resource attribute-id '%s' found with value %s\n", logstr, attribute_id, value);
                }
            }
        }
    }
    return 0;
}



/************************************************
Function:    destroyRequestSubject
Parameters:  xacml_profile_request_t * xacml_profile_request
Description:
             This function will destroy and free the Subject information details
             from the request, which was set by the extractRequestSubject() function
             to a xacml_profile_request_t
Return:
          0 : good
         !0 : bad 
************************************************/
int destroyRequestSubject (xacml_profile_request_t * xacml_profile_request)
{
    size_t i     = 0;


    if (xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_id.subject_x509_id)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_id.subject_x509_id);
        xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_id.subject_x509_id = NULL;
    }
    if (xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_issuer.subject_x509_issuer)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_issuer.subject_x509_issuer);
        xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_issuer.subject_x509_issuer = NULL; 
    } 
    if (xacml_profile_request -> xacml_profile_subject.xacml_cert_chain.cert_chain)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_cert_chain.cert_chain);
        xacml_profile_request -> xacml_profile_subject.xacml_cert_chain.cert_chain = NULL;
    }

    if (xacml_profile_request -> xacml_profile_subject.xacml_subject_primary_voms_fqan.subject_primary_voms_fqan)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_subject_primary_voms_fqan.subject_primary_voms_fqan);
        xacml_profile_request -> xacml_profile_subject.xacml_subject_primary_voms_fqan.subject_primary_voms_fqan = NULL;
    }

    /* The CA_serial number can appear multiple times, but is an integer which doesn't need the loop-ed free() construct */ 
    free(xacml_profile_request -> xacml_profile_subject.xacml_ca_serial_number);
    xacml_profile_request -> xacml_profile_subject.xacml_ca_serial_number = NULL;

    /* Freeing multi-valued elements */
    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid_cnt; i++)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid[i].ca_policy_oid);
        xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid[i].ca_policy_oid = NULL;
    }
    if (xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid);
        xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid = NULL;
    }


    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_subject_vo_cnt; i++)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_subject_vo[i].subject_vo);
        xacml_profile_request -> xacml_profile_subject.xacml_subject_vo[i].subject_vo = NULL; 
    } 
    if (xacml_profile_request -> xacml_profile_subject.xacml_subject_vo)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_subject_vo);
        xacml_profile_request -> xacml_profile_subject.xacml_subject_vo = NULL;
    }

    
    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject_cnt; i++)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject[i].voms_signing_subject);
        xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject[i].voms_signing_subject = NULL;
    }
    if (xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject);
        xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject = NULL;
    }


    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer_cnt; i++)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer[i].voms_signing_issuer);
        xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer[i].voms_signing_issuer = NULL;
    }
    if (xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer);
        xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer = NULL;
    }


    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port_cnt; i++)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port[i].voms_dns_port);
        xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port[i].voms_dns_port = NULL;
    }
    if (xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port);
        xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port = NULL;
    }

    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan_cnt; i++)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan[i].subject_voms_fqan);
        xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan[i].subject_voms_fqan = NULL;
    }
    if (xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan);
        xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan = NULL;
    }


    return 0;
}


/************************************************
Function:    destroyRequestAction
Parameters:  xacml_profile_request_t * xacml_profile_request
Description:
             This function will destroy and free the Action information details
             from the request, which was set by the extractRequestAction() function
             to a xacml_profile_request_t
Return:
          0 : good
         !0 : bad 
************************************************/
int destroyRequestAction (xacml_profile_request_t * xacml_profile_request)
{
    if (xacml_profile_request -> xacml_profile_action.xacml_action_rsl_string.action_rsl_string)
    {
        free (xacml_profile_request -> xacml_profile_action.xacml_action_rsl_string.action_rsl_string);
    }

    return 0;
}



/************************************************
Function:    destroyRequestResource
Parameters:  xacml_profile_request_t * xacml_profile_request
Description:
             This function will destroy and free the Resource information details
             from the request, which was set by the extractRequestResource() function
             to a xacml_profile_request_t
Return:
          0 : good
         !0 : bad 
************************************************/
int destroyRequestResource (xacml_profile_request_t * xacml_profile_request)
{
    if (xacml_profile_request -> xacml_profile_resource.xacml_resource_x509_id.resource_x509_id)
    {
        free (xacml_profile_request -> xacml_profile_resource.xacml_resource_x509_id.resource_x509_id);
    }
    if (xacml_profile_request -> xacml_profile_resource.xacml_resource_x509_issuer.resource_x509_issuer)
    {
        free (xacml_profile_request -> xacml_profile_resource.xacml_resource_x509_issuer.resource_x509_issuer);
    }
    if (xacml_profile_request -> xacml_profile_resource.xacml_resource_dns_name.resource_dns_name)
    {
        free (xacml_profile_request -> xacml_profile_resource.xacml_resource_dns_name.resource_dns_name);
    }

    return 0;
}


/************************************************
Function:    constructFailureResponse
Parameters:  xacml_response_t * response
Description:
             The XACML response message to indicate a failure is constructed here.
************************************************/
int constructFailureResponse (xacml_response_t * response)
{
    xacml_response_set_saml_status_code  (*response, SAML_STATUS_AuthnFailed);
    xacml_response_set_xacml_status_code (*response, XACML_STATUS_ok);
    xacml_response_set_xacml_decision    (*response, XACML_DECISION_Deny);

    return 0;
}



/************************************************
Function:    constructResponse
Parameters:  xacml_response_t * response
             uid_t uid
             gid_t pgid
             gid_t sgids[]
             int   sgid_cnt
Description:
             The XACML response message is constructed here.
             This implementation will use the Unix UID, Primary GID and
             multiple Secondary GIDs as input to contruct the obligations and
             its attributes. 
************************************************/
int constructResponse (xacml_response_t * response,
                       uid_t uid,
                       gid_t pgid,
                       gid_t sgids[],
                       int   sgid_cnt)
{
    /* Mapping Information Translated */
    char *  str_uid       = NULL;
    char *  str_pgid      = NULL;
    char ** str_sgids     = NULL;
    size_t  str_sgids_cnt = 0;
    size_t  i             = 0;
    /* Mapping Information Translated */

    xacml_obligation_t  obligation_uidgid                = NULL;
    xacml_obligation_t  obligation_secgids               = NULL;
#ifdef USE_AFSTOKEN
    xacml_obligation_t  obligation_afstoken              = NULL;
#endif /* USE_AFSTOKEN */

#ifdef STAKEHOLDER_OSG
    xacml_obligation_t  obligation_username              = NULL;
    xacml_obligation_t  obligation_RootAndHomePaths      = NULL;
    xacml_obligation_t  obligation_StorageAccessPriority = NULL;
    xacml_obligation_t  obligation_AccessPermissions     = NULL;
#endif /* STAKEHOLDER_OSG */


    /* transform! - 2^32 uses 10 bytes */
    /* str_uid = malloc (sizeof (char) * 11); */
    str_uid = calloc (11, sizeof (char));
    snprintf (str_uid, 11, "%d", uid);

    /* str_pgid = malloc (sizeof (char) * 11); */
    str_pgid = calloc (11, sizeof (char));
    snprintf (str_pgid, 11, "%d", pgid);

    if (sgid_cnt>0) {
	str_sgids_cnt = (size_t)sgid_cnt;
	/* str_sgids = malloc (sizeof (int) * str_sgids_cnt); */
	str_sgids = calloc (str_sgids_cnt, sizeof (char **));
	for (i = 0; i < str_sgids_cnt; i++)
	{
	    /* str_sgids[i] = malloc (sizeof (char) * 11); */
	    str_sgids[i] = calloc (11, sizeof (char));
	    snprintf (str_sgids[i], 11, "%d", sgids[i]); 
	}
	/* transform! - 2^32 uses 10 bytes */
    } else  {
	str_sgids_cnt = 0U;
/*	str_sgids = calloc(1,1);*/
	str_sgids = NULL;
    }





    /*********** B: Obligation UIDGID ***********/
    xacml_obligation_init(&obligation_uidgid,
                          XACML_ATTR_PROFILE_NS_URL_OBLIGATION_UIDGID,
                          XACML_EFFECT_Permit);
    /* uid */
    xacml_obligation_add_attribute(obligation_uidgid,
                                   XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_UID,
                                   XACML_DATATYPE_INTEGER,
                                   str_uid);

    /* gid */
    xacml_obligation_add_attribute(obligation_uidgid,
                                   XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_GID,
                                   XACML_DATATYPE_INTEGER,
                                   str_pgid);

    xacml_response_add_obligation(*response, obligation_uidgid);
    xacml_obligation_destroy(obligation_uidgid);
    /*********** E: Obligation UIDGID ***********/


    /*********** B: Obligation SecondaryGIDs ***********/
    xacml_obligation_init(&obligation_secgids,
                          XACML_ATTR_PROFILE_NS_URL_OBLIGATION_SecondaryGIDs,
                          XACML_EFFECT_Permit);

    for (i = 0; i < str_sgids_cnt; i++)
    {
        xacml_obligation_add_attribute(obligation_secgids,
                                       XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_GID,
                                       XACML_DATATYPE_INTEGER,
                                       str_sgids[i]);
    }

    xacml_response_add_obligation(*response, obligation_secgids);
    xacml_obligation_destroy(obligation_secgids);
    /*********** E: Obligation SecondaryGIDs ***********/



    xacml_response_set_saml_status_code  (*response, SAML_STATUS_Success);
    xacml_response_set_xacml_status_code (*response, XACML_STATUS_ok);
    xacml_response_set_xacml_decision    (*response, XACML_DECISION_Permit);



    /* Free stage */
    free(str_uid);
    free(str_pgid);

    for (i = 0; i < str_sgids_cnt; i++)
    {
        free(str_sgids[i]);
    }
    /* Free if at least one sgid was malloc'ed */
    if (str_sgids_cnt)
        free(str_sgids);
    /* Free stage */ 

    return 0;
}


/************************************************
Function:    printRecvRequest
Description:
             Prints the parsed and fetched data from the XACML request
************************************************/
void printRecvRequest (xacml_profile_request_t * xacml_profile_request)
{
    size_t i;
    const int ll = 5;


    scas_log_debug (ll, "--- Received SAML/XACML Attributes ----------------------------------------------------------\n");
    if (xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_id.subject_x509_id)
        scas_log_debug (ll, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_SUBJECT_X509_ID, xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_id.subject_x509_id);
    if (xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_issuer.subject_x509_issuer)
        scas_log_debug (ll, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_SUBJECT_X509_ISSUER, xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_issuer.subject_x509_issuer);
    if (xacml_profile_request -> xacml_profile_subject.xacml_subject_certificate_serial_number.certificate_serial_number)
        scas_log_debug (ll, "   %s: %d\n", XACML_ATTR_PROFILE_NS_URL_SUBJECT_CERTIFICATE_SERIAL_NUMBER, 
                                     xacml_profile_request -> xacml_profile_subject.xacml_subject_certificate_serial_number.certificate_serial_number);
    if (xacml_profile_request -> xacml_profile_subject.xacml_cert_chain.cert_chain)
        scas_log_debug (ll, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_CERT_CHAIN, xacml_profile_request -> xacml_profile_subject.xacml_cert_chain.cert_chain);
    if (xacml_profile_request -> xacml_profile_subject.xacml_subject_primary_voms_fqan.subject_primary_voms_fqan)
        scas_log_debug (ll, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_SUBJECT_VOMS_PRIMARY_FQAN, xacml_profile_request -> xacml_profile_subject.xacml_subject_primary_voms_fqan.subject_primary_voms_fqan);

    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_ca_serial_number_cnt; i++)
        scas_log_debug (ll, "   %s: [%d] %d\n", XACML_ATTR_PROFILE_NS_URL_SUBJECT_CA_SERIAL_NUMBER, i, xacml_profile_request -> xacml_profile_subject.xacml_ca_serial_number[i].ca_serial_number);
    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid_cnt; i++)
        scas_log_debug (ll, "   %s: [%d] %s\n", XACML_ATTR_PROFILE_NS_URL_CA_POLICY_OID, i, xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid[i].ca_policy_oid);
    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_subject_vo_cnt; i++)
        scas_log_debug (ll, "   %s: [%d] %s\n", XACML_ATTR_PROFILE_NS_URL_VO, i, xacml_profile_request -> xacml_profile_subject.xacml_subject_vo[i].subject_vo);
    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject_cnt; i++)
        scas_log_debug (ll, "   %s: [%d] %s\n", XACML_ATTR_PROFILE_NS_URL_VOMS_SIGNING_SUBJECT, i, xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject[i].voms_signing_subject);
    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer_cnt; i++)
        scas_log_debug (ll, "   %s: [%d] %s\n", XACML_ATTR_PROFILE_NS_URL_VOMS_SIGNING_ISSUER, i, xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer[i].voms_signing_issuer);
    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port_cnt; i++)
        scas_log_debug (ll, "   %s: [%d] %s\n", XACML_ATTR_PROFILE_NS_URL_VOMS_DNS_PORT, i, xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port[i].voms_dns_port);
    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan_cnt; i++)
        scas_log_debug (ll, "   %s: [%d] %s\n", XACML_ATTR_PROFILE_NS_URL_SUBJECT_VOMS_FQAN, i, xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan[i].subject_voms_fqan);
        /* Multiple elements */

    /* Action */
    if (xacml_profile_request -> xacml_profile_action.xacml_action_job_type == ACTION_QUEUE)
        scas_log_debug (ll, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_ID, XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_QUEUE);
    else if (xacml_profile_request -> xacml_profile_action.xacml_action_job_type == ACTION_EXECUTE_NOW)
        scas_log_debug (ll, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_ID, XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_EXEC_NOW);
    else if (xacml_profile_request -> xacml_profile_action.xacml_action_job_type == ACTION_ACCESSFILE)
        scas_log_debug (ll, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_ID, XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_ACCESS_FILE);

    /* Resource */
    if (xacml_profile_request -> xacml_profile_resource.xacml_resource_node_type == RESOURCE_CE)
        scas_log_debug (ll, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_ID, XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_CE);
    else if (xacml_profile_request -> xacml_profile_resource.xacml_resource_node_type == RESOURCE_SE)
        scas_log_debug (ll, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_ID, XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_SE);
    else if (xacml_profile_request -> xacml_profile_resource.xacml_resource_node_type == RESOURCE_WN)
        scas_log_debug (ll, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_ID, XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_WN);
    else if (xacml_profile_request -> xacml_profile_resource.xacml_resource_node_type == RESOURCE_RB)
        scas_log_debug (ll, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_ID, XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_RB);

    if (xacml_profile_request -> xacml_profile_resource.xacml_resource_x509_id.resource_x509_id)
        scas_log_debug (ll, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_X509_ID, xacml_profile_request -> xacml_profile_resource.xacml_resource_x509_id.resource_x509_id);
    if (xacml_profile_request -> xacml_profile_resource.xacml_resource_x509_issuer.resource_x509_issuer)
        scas_log_debug (ll, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_X509_ISSUER, xacml_profile_request -> xacml_profile_resource.xacml_resource_x509_issuer.resource_x509_issuer);
    if (xacml_profile_request -> xacml_profile_resource.xacml_resource_dns_name.resource_dns_name)
        scas_log_debug (ll, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_RESOURCE_DNS_HOST_NAME, xacml_profile_request -> xacml_profile_resource.xacml_resource_dns_name.resource_dns_name);

    /* Environment */
    /* Todo */
    scas_log_debug (ll, "--- Received SAML/XACML Attributes ----------------------------------------------------------\n");
}


/************************************************
Function:    xacml_authorize
Description:
             This is the function that is called to get the request and construct the response.
             All the elements from the request are made available by the saml2-xacml2 library and
             it allows to call out to an external facility to get more information that would 
             need to go into the response.
Note:        This is the function that is registered in the SAML2-XACML2 library
             from my server.c example file
Return:


************************************************/
int
xacml_authorize(
    void *                              handler_arg,
    const xacml_request_t               request,
    xacml_response_t                    response)
{
    char * logstr = "xacml_authorize()";
    int rc = 0;
    int i = 0;
/*    xacml_status_code_t my_decision = XACML_STATUS_ok;*/
    lcmaps_account_info_t paccount_info;
    /* time_t time_stamp = time(NULL); */


    xacml_response_set_issuer(response, "SCAS");
    lcmaps_account_info_init (&paccount_info);


    /* XACML Interoperability Profile Request */
    xacml_profile_request_t * xacml_profile_request;
    /* xacml_profile_request = malloc (sizeof (xacml_profile_request_t)); */
    xacml_profile_request = calloc (1, sizeof (xacml_profile_request_t));
    if (!xacml_profile_request)
    {
        scas_log(LOG_ERR, "%s: Out of memory. Couldn't allocate %d bytes of memory\n", logstr, sizeof (xacml_profile_request_t));
        return -1;
    }
    /* bzero (xacml_profile_request, sizeof (xacml_profile_request_t)); */
    /* XACML Interoperability Profile Request */



    /* Extract information and push it to the internal structs for direct access to the message elements */
    if (extractRequestSubject (request, xacml_profile_request) < 0)
    {
        scas_log(LOG_ERR, "%s: Failure in extractRequestSubject()\n", logstr);
/*        my_decision = XACML_STATUS_processing_error;*/
        goto post_everything_cleanup;
    }

    if (extractRequestAction (request, xacml_profile_request) < 0)
    {
        scas_log(LOG_ERR, "%s: Failure in extractRequestAction()\n", logstr);
/*        my_decision = XACML_STATUS_processing_error;*/
        goto post_everything_cleanup;
    }

    if (extractRequestResource (request, xacml_profile_request) < 0)
    {
        scas_log(LOG_ERR, "%s: Failure in extractRequestResource()\n", logstr);
/*        my_decision = XACML_STATUS_processing_error;*/
        goto post_everything_cleanup;
    }
    /* Extract information and push it to the internal structs for direct access to the message elements */

    printRecvRequest (xacml_profile_request);   


#if 0
    /* Disabled LCAS ban0ing on the pilot job subject for the moment */
    /* FIX LCAS ! */
    if ((rc = execLCAS (xacml_profile_request)))
    {
        scas_log(LOG_ERR, "%s: LCAS Authorization failed!\n", logstr);

        /* Construct failure response */
        constructFailureResponse (&response);

        goto post_everything_cleanup;
    } 
#endif



    /*********** Apply mutex ***********/
    pthread_mutex_lock(&mutex);
    /*********** Apply mutex ***********/

    rc = execLCMAPS (xacml_profile_request, &paccount_info);
    if (rc)
    {
        scas_log(LOG_ERR, "%s: LCMAPS failed with the return code: %d\n", logstr, rc);

        /*********** Release Mutex ***********/
        pthread_mutex_unlock(&mutex);
        /*********** Release Mutex ***********/

        /* Construct failure response */
        constructFailureResponse (&response);

        lcmaps_account_info_clean(&paccount_info);
        goto post_everything_cleanup;
    }

    /*********** Release Mutex ***********/
    pthread_mutex_unlock(&mutex);
    /*********** Release Mutex ***********/


    /* Construct log message of LCMAPS's mapping results */
    scas_log_debug (3, "%s: LCMAPS returned: UID: '%d' GID: '%d'\n", 
                       logstr, 
                       paccount_info.uid, 
                       paccount_info.pgid_list[0]);
    for (i = 0; i < paccount_info.nsgid; i++)
        scas_log_debug (3, "%s:                             SGID: '%d'\n", 
                           logstr, 
                           paccount_info.sgid_list[i]); 

    /* Construct full response, build success statement in response */
    constructResponse (&response, 
                       paccount_info.uid, 
                       paccount_info.pgid_list[0], 
                       paccount_info.sgid_list, paccount_info.nsgid);

    lcmaps_account_info_clean(&paccount_info);


post_everything_cleanup:
    /* Destroy and free Profile Request */
    destroyRequestSubject (xacml_profile_request);
    destroyRequestAction (xacml_profile_request);
    destroyRequestResource (xacml_profile_request);
    free(xacml_profile_request);

    /* Always '0' */
    return 0;

}
/********** xacml_authorize() **********/


